/*

Copyright (C) 1996 John W. Eaton

This file is part of Octave.

Octave is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2, or (at your option) any
later version.

Octave is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Octave; see the file COPYING.  If not, write to the Free
Software Foundation, 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <string>

#include<iostream.h>

#include "defun-dld.h"
#include "dirfns.h"
#include "error.h"
#include "help.h"
#include "oct-map.h"
#include "systime.h"
#include "ov.h"
#include "oct-obj.h"
#include "utils.h"
#include "oct-env.h"

#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <errno.h>
#include <netdb.h>
#include <unistd.h>

#define BUFF_SIZE SSIZE_MAX

// COMM

DEFUN_DLD (reval, args, ,
  "reval (commands,sockets)\n\
\n\
Evaluate 'commands' at the remote hosts specified by the matrix 'sockets'.")
{
  octave_value retval;

  if(args.length () ==2)
    {
      int sock,row=0,col=0,nsock=0,i,j;//,error_state;
      octave_value val=args(0);
      Matrix sock_m=args(1).matrix_value();
      charMatrix cm=val.char_matrix_value();
      char comm[256];
      
      nsock=sock_m.rows();

      row=val.rows();
      col=val.columns();
      for(i=0;i<nsock;i++){
	sock=(int)sock_m.data()[i+nsock];
	if(sock!=0){
	  for(j=0;j<row;j++){
	    strncpy(comm,(cm.extract(j,0,j,col-1).data()),col);
	    comm[col]='\n';
	    comm[col+1]='\0';
	    write(sock,&col,sizeof(col));
	    write(sock,comm,col);

	    // Blocking Execution
//	    read(sock,&error_state,sizeof(int));
//	    if(error_state){
//	      cout << "\nError occurred on " << sock<<endl;
//	      jump_to_top_level ();
//	      panic_impossible ();
//	    }

	  }
	}
      }
    }
  else
    print_usage ("reval");

  return retval;

}



DEFUN_DLD (send, args, ,
  "send (X,sockets)\n\
\n\
Send the variable 'x' to the computers specified by matrix 'sockets'.")
{
  octave_value retval;
  
  if(args.length () ==2)
    {
      octave_value val=args(0);
      Matrix sock_m=args(1).matrix_value();
      int sock,i;
      int nhosts=sock_m.rows();
      int type_id=val.type_id();
      
      if(val.is_real_matrix() && !(val.is_char_matrix()))
	{
	  Matrix m=val.matrix_value();
	  int row=m.rows();
	  int col=m.columns();
	  int length=0;
	  
	  for (i=0;i<nhosts;i++){
	    sock=(int)sock_m.data()[i];
	    if(sock!=0){
	      write(sock,&type_id,sizeof(type_id));
	      write(sock,&row,sizeof(row));
	      write(sock,&col,sizeof(col));
	      length=sizeof(double)*col*row;
	      errno=0;
	      write(sock,m.data(),length);
	    }
	  }
	  
	}
      else if(val.is_real_scalar())
	{
	  double d=val.double_value();
	  int length=sizeof(d);
	  
	  for (i=0;i<nhosts;i++){
	    sock=(int)sock_m.data()[i];
	    if(sock!=0){
	      write(sock,&type_id,sizeof(type_id));
	      write(sock,&d,length);
	    }
	  }
	}
      else if(val.is_complex_matrix())
	{
	  ComplexMatrix m=val.complex_matrix_value();
	  int row=m.rows();
	  int col=m.columns();
	  int length=0;
	  
	  for (i=0;i<nhosts;i++){
	    sock=(int)sock_m.data()[i];
	    if(sock!=0){
	      write(sock,&type_id,sizeof(type_id));
	      write(sock,&row,sizeof(row));
	      write(sock,&col,sizeof(col));
	      length=sizeof(Complex)*col*row;
	      write(sock,m.data(),length);
	    }
	  }
	}
      else if(val.is_complex_scalar())
	{
	  Complex cx=val.complex_value();
	  int length=sizeof(cx);
	  
	  for (i=0;i<nhosts;i++){
	    sock=(int)sock_m.data()[i];
	    if(sock!=0){
	      write(sock,&type_id,sizeof(type_id));
	      write(sock,&cx,length);
	    }
	  }
	}
      else if(val.is_char_matrix())
	{
	  int row=val.rows();
	  int col=val.columns();
	  int length=sizeof(char)*row*col;
	  charMatrix cmx=val.char_matrix_value();
	  
	  for (i=0;i<nhosts;i++){
	    sock=(int)sock_m.data()[i];
	    if(sock!=0){
	      write(sock,&type_id,sizeof(type_id));
	      write(sock,&row,sizeof(row));
	      write(sock,&col,sizeof(col));
	      write(sock,cmx.data(),length);
	    }
	  }
	}
      else if(val.is_map())
	{
	  Octave_map map=val.map_value();
	  octave_value_list ov_list;
	  Pix pix=map.first();
	  int i,length=map.length(),key_len=0;
	  string key=map.key(pix);
	  
  	  for (i=0;i<nhosts;i++){
  	    sock=(int)sock_m.data()[i];
	    ov_list(1)=octave_value(sock_m.row(i));
	    if(sock!=0){
	      write(sock,&type_id,sizeof(type_id));
	      write(sock,&length,sizeof(length));
	      for(i=0;i<length;i++){
		key_len=key.length();
		write(sock,&key_len,sizeof(key_len));
		ov_list(0)=map[key];
		write(sock,key.c_str(), key_len);
		Fsend(ov_list,0);
		map.next(pix);
		if(pix!=NULL)
		  key=map.key(pix);
	      }
	    }
	  }
	}
    }
  else
    print_usage ("send");
  
  return retval;
  
}


DEFUN_DLD (connect, args, ,
  "connect (hosts)\n\
\n\
Connect hosts and return sockets.")
{

  int sock=0,col=0,row=0,i,j,len;
  double *sock_v=0;
  if (args.length () == 1)
    {
      int pid=0;
      struct sockaddr_in *addr;
      struct hostent *he;
      octave_value hosts=args(0);
      charMatrix cm=hosts.char_matrix_value();
      char *host,*pt,myname[16];
      
      errno=0;
      gethostname(myname,15);
      row= cm.rows();
      col= cm.columns();
      cm= cm.transpose(); 
      
      sock_v=(double *)calloc(2,row*sizeof(double));
       

      for(i=1;i<row;i++){
	host=(char *)calloc(1,col+1);
	strncpy(host,&cm.data()[col*i],col);
	pt=strchr(host,' ');
	if(pt==NULL)
	  host[col]='\0';
	else
	  pt='\0';
	
	sock=socket(PF_INET,SOCK_STREAM,0);
	if(sock==-1){
	  perror("socket : ");
	  exit(-1);
	}
	
	addr=(struct sockaddr_in *) calloc(1,sizeof(struct sockaddr_in));
	
	addr->sin_family=AF_INET;
	addr->sin_port=htons(12502);
	he=gethostbyname(host);
	memcpy(&addr->sin_addr,he->h_addr_list[0],he->h_length);
	
	for(j=0;j<10;j++){
	  if(connect(sock,(struct sockaddr *)addr,sizeof(*addr))==0){
	    break;
	  }else if(errno!=ECONNREFUSED){
	    perror("connect : ");
	    exit(-1);
	  }else {
	    usleep(5000);
	  }
	}
	
	sock_v[i+row]=sock;
	
	free(addr);
	free(host);

	int num_nodes=row-1;

	pid=getpid();
	write(sock,&num_nodes,sizeof(int));
	write(sock,&i,sizeof(int));
	write(sock,&pid,sizeof(int));
	host=(char *)calloc(128,sizeof(char));
	for(j=0;j<row;j++){
	  strncpy(host,&cm.data()[col*j],col);
	  pt=strchr(host,' ');
	  if(pt==NULL)
	    host[col]='\0';
	  else
	    pt='\0';
	  len=strlen(host)+1;
	  write(sock,&len,sizeof(int));
	  write(sock,host,len);
	}
	free(host);
	int comm_len;
	char *comm;
       	std::string directory = octave_env::getcwd ();
	comm=directory.c_str();
	comm_len=directory.length();
	write(sock,&comm_len,sizeof(int));
	write(sock,directory.c_str(),comm_len);
      }      
      usleep(100);

      for(i=1;i<row;i++){
	
	host=(char *)calloc(1,col+1);
	
	strncpy(host,&cm.data()[col*i],col);
	pt=strchr(host,' ');
	if(pt==NULL)
	  host[col]='\0';
	else
	  pt='\0';
	
	sock=socket(PF_INET,SOCK_STREAM,0);
	if(sock==-1){
	  perror("socket : ");
	  exit(-1);
	}
	
	addr=(struct sockaddr_in *) calloc(1,sizeof(struct sockaddr_in));
	
	addr->sin_family=AF_INET;
	addr->sin_port=htons(12501);
	he=gethostbyname(host);
	memcpy(&addr->sin_addr,he->h_addr_list[0],he->h_length);
	while(1){
	  for(j=0;j<10;j++){
	    if(connect(sock,(struct sockaddr *)addr,sizeof(*addr))==0){
	      break;
	    }else if(errno!=ECONNREFUSED){
	      perror("connect : ");
	      exit(-1);
	    }else {
	      usleep(5000);
	    }
	  }
	  
	  int bufsize=262144;
	  socklen_t ol;
	  ol=sizeof(bufsize);
	  setsockopt(sock,SOL_SOCKET,SO_SNDBUF,&bufsize,ol);
	  setsockopt(sock,SOL_SOCKET,SO_RCVBUF,&bufsize,ol);
	  bufsize=1;
	  ol=sizeof(bufsize);
	  setsockopt(sock,SOL_SOCKET,SO_REUSEADDR,&bufsize,ol);
	  

	  int len=0,result=0;;
	  //send pppid
	  write(sock,&pid,sizeof(int));
	  //send name size
	  strncpy(myname,cm.data(),col);
	  myname[col]='\0';
	  len=strlen(myname);
	  write(sock,&len,sizeof(int));
	  //send name
	  write(sock,myname,len+1);
	  //recv result code
	  read(sock,&result,sizeof(int));
	  if(result==0){
	    sock_v[i]=sock;
	    break;
	  }else{
	    close(sock);
	  }
	}
	
	free(addr);
	free(host);
      }

      char lf='\n';
      for(i=1;i<row;i++){
	write((int)sock_v[i+row],&lf,sizeof(char));
	//	cout << i+row <<endl;
      }
    }
  else
    print_usage ("connect");

  Matrix mx(row,2);
  double *tmp =mx.fortran_vec();
  for (i=0;i<2*row;i++)
    tmp[i]=sock_v[i];
  octave_value retval(mx);

  return retval;
}



DEFUN_DLD (close, args, ,
  "close (sockets)\n\
\n\
Close sockets")
{
  octave_value retval;
  errno=0;

  if(args.length () == 1)
    {
      int i,nsock=0,sock;
      nsock=args(0).matrix_value().rows()*2;
      
      for(i=nsock-1;i>0;i--){
	sock=(int)args(0).matrix_value().data()[i];
	if(sock!=0){
	  if(close(sock)!=0)
	    perror("close");
	}
      }
    }
  else
    print_usage ("close");

  return retval;

}



DEFUN_DLD (recv, args, ,
  "recv (socket)\n\
\n\
Receive a variable from the computer specified by the row vector 'socket'.")
{
  octave_value retval;
  int type_id,sock;
  
  if(args.length () == 1)
    {
      Matrix m=args(0).matrix_value();
      
      sock=(int) m.data()[0];
      
      read(sock,&type_id,sizeof(type_id));
      if(type_id==3)
	{
	  int col=0,row=0,length=0,count=0,r_len=0;
	  read(sock,&row,sizeof(row));
	  read(sock,&col,sizeof(col));
	  length=sizeof(double)*row*col;
	  Matrix m(row,col);
	  double* tmp = m.fortran_vec();
	  errno=0;
	  r_len=BUFF_SIZE;
	  while(count <length){
	    if((length-count) < BUFF_SIZE)
	      r_len=length-count;
	    count +=read(sock,(double *)((int)tmp+count),r_len);
	  }
	  retval= octave_value(m);
	}
      else if(type_id==1)
	{
	  double d;
	  
	  read(sock,&d,sizeof(d));
	  retval= octave_value(d);
	  
	}
      else if(type_id==4)
	{
	  int col=0,row=0,length=0,count=0,r_len=0;
	  
	  read(sock,&row,sizeof(row));
	  read(sock,&col,sizeof(col));
	  length=sizeof(Complex)*row*col;
	  ComplexMatrix cm(row,col);
	  Complex* tmp = cm.fortran_vec();
	  errno=0;
	  r_len=BUFF_SIZE;
	  while(count <length){
	    if((length-count) < BUFF_SIZE)
	      r_len=length-count;
	    count +=read(sock,(Complex *)((int)tmp+count),r_len);
	  }
	  retval= octave_value(cm);
	  
	}
      else if(type_id==2)
	{
	  Complex cx;
	  
	  read(sock,&cx,sizeof(cx));
	  retval= octave_value(cx);
	  
	}
      else if(type_id==7)
	{
	  int col=0,row=0,length=0,count=0,r_len=0;
	  
	  read(sock,&row,sizeof(row));
	  read(sock,&col,sizeof(col));
	  length=sizeof(char)*row*col;
	  charMatrix cmx(row,col);
	  char* str = cmx.fortran_vec();
	  errno=0;
	  r_len=BUFF_SIZE;
	  while(count <length){
	    if((length-count) < BUFF_SIZE)
	      r_len=length-count;
	    count +=read(sock,(char *)((int)str+count),r_len);
	  }
	  retval= octave_value(cmx,1);
	  
	}
      else if(type_id==10)
	{
	  int i,length=0,key_len=0;
	  Octave_map m;
	  octave_value_list ov_list;
	  char *key;
	  
	  read(sock,&length,sizeof(length));
	  for(i=0;i<length;i++){
	    read(sock,&key_len,sizeof(key_len));
	    key = (char *)calloc(sizeof(char),key_len+1);
	    read(sock,key,key_len);
	    ov_list=Frecv(args(0),0);
	    m [key]=ov_list(0); 
	  }
	  retval=octave_value(m);
	}
    }
  else
    print_usage ("recv");

  return retval;

}


DEFUN_DLD (obj, args, ,
    "obj (X)\n\
 \n\
  obj type")
{
  octave_value val=args(0),retval;
  cout << val.matrix_value().row(1) <<'\n';
  cout << "args.length: " << args.length() <<'\n';
  cout << "row: " << val.rows() <<'\n';
  cout << "columns: " << val.columns()  <<'\n';
  cout << "is_defined: " << val.is_defined()  <<'\n';
  cout << "is_real_scalar: " << val.is_real_scalar()  <<'\n';
  cout << "is_real_matrix: " << val.is_real_matrix() <<'\n';
  cout << "is_complex_scalar: " << val.is_complex_scalar()  <<'\n';
  cout << "is_complex_matrix: " << val.is_complex_matrix() <<'\n';
  cout << "is_char_matrix: " << val.is_char_matrix() <<'\n';
  cout << "is_string: " << val.is_string()  <<'\n';
  cout << "is_range: " <<val.is_range()  <<'\n';
  cout << "is_map: " << val.is_map()  <<'\n';
  cout << "is_magic_colon: " << val.is_magic_colon()  <<'\n';
  cout << "is_all_va_args: " << val.is_all_va_args() <<'\n';
  cout << "is_real_type: " << val.is_real_type()  <<'\n';
  cout << "is_complex_type: " << val.is_complex_type()  <<'\n';
  cout << "is_scalar_type: " <<val.is_scalar_type()  <<'\n';
  cout << "is_matrix_type: " <<val.is_matrix_type() <<'\n';
  cout << "is_numeric_type: " << val.is_numeric_type()  <<'\n';
  //  cout << "valid_as_scalar_index: " << val.valid_as_scalar_index() <<'\n';
  //cout << "valid_as_zero_index: " << val.valid_as_zero_index() <<'\n';
  cout << "is_true: " << val.is_true() <<'\n';
  cout << "is_empty: " << val.is_empty() <<'\n';
  cout << "is_zero_by_zero: " << val.is_zero_by_zero() <<'\n';
  cout << "type_id: " << val.type_id()  <<'\n';
  cout << "type_name: " << val.type_name()  <<'\n';
//    cout << val.matrix_value() <<'\n';
//    //  cout << val.matrix_value().size()<< '\n';
//    cout << val.matrix_value().length() << '\n';
//    retval=val.all();
  if(val.is_map()){
    Octave_map map=val.map_value(),map2;
    Pix pix=map.first();
    map.next(pix);
    string key=map.key(pix);
    cout << key <<endl;
    map2[key]=map[key];
    //    retval=key;
    cout << map.length() <<endl;
    retval=map2;
  }

//    //  cout << sizeof(*m) << '\n' ;
  //  cout << val.char_matrix_value() <<endl;
  return retval;
}

/*
;;; Local Variables: ***
;;; mode: C++ ***
;;; End: ***
*/
